import base64
import copy
from abcli import env
from enum import Enum, auto
import os
import os.path
from abcli import file
from . import NAME
from abcli.logger import logger


class MetadataSourceType(Enum):
    FILENAME = auto()
    OBJECT = auto()
    PATH = auto()

    def filename(self, source: str, filename="metadata.yaml"):
        if self == MetadataSourceType.FILENAME:
            return source

        if self == MetadataSourceType.OBJECT:
            from abcli.modules import objects

            return os.path.join(
                objects.object_path(source),
                filename,
            )

        if self == MetadataSourceType.PATH:
            return os.path.join(
                env.abcli_object_path if source == "." else source,
                filename,
            )

        return f"unknown-source-type:{self.name.lower()}"


def get(
    key,
    default="",
    source=".",
    source_type: MetadataSourceType = MetadataSourceType.FILENAME,
    dict_keys: bool = False,
    dict_values: bool = False,
    filename: str = "metadata.yaml",
):
    success, metadata = file.load_yaml(source_type.filename(source, filename))
    if not success:
        return default

    try:
        output = metadata
        for key_ in [key_ for key_ in key.split(".") if key_]:
            output = output.get(key_, default)

        if dict_keys:
            output = list(output.keys())
        elif dict_values:
            output = list(output.values())

        return output
    except Exception as e:
        return type(e).__name__


def post(
    key,
    value,
    filename: str = "metadata.yaml",
    source=".",
    source_type: MetadataSourceType = MetadataSourceType.FILENAME,
    is_base64_encoded=False,
    verbose: bool = False,
) -> bool:
    if is_base64_encoded:
        value = str(base64.b64decode(value))

    filename = source_type.filename(source, filename)

    _, metadata = file.load_yaml(filename, civilized=True)

    metadata[key] = copy.deepcopy(value)

    logger.info(
        "{}.post[{}]: {}{}".format(
            NAME,
            filename,
            key,
            f"={value}" if verbose else "",
        )
    )

    return file.save_yaml(filename, metadata)



import cv2
import math
from typing import List
import numpy as np
from .text import render_text
from . import NAME
from abcli import string
from functools import reduce
from abcli.logger import logger


def add_sidebar(image, lines, images=[], line_length=28):
    logger.debug(
        "{}.add_sidebar({},{},{})".format(
            NAME,
            string.pretty_shape_of_matrix(image),
            "|".join(lines),
            len(images),
        )
    )

    if not image.shape or not lines:
        return image

    if line_length == -1:
        line_length = max([len(line) for line in lines])

    image_width = int(image.shape[1] / 7)

    sidebar = np.concatenate(
        [
            image_
            for _, image_ in sorted(
                zip(
                    lines,
                    [
                        render_text(
                            text=line.ljust(line_length),
                            color_depth=image.shape[2],
                            image_width=image_width,
                        )
                        for line in lines
                    ],
                )
            )
        ]
        + [
            cv2.resize(
                image_,
                (
                    image_width,
                    int(image_.shape[1] / image_.shape[0] * image_width),
                ),
            )
            for image_ in images
        ],
        axis=0,
    )

    if sidebar.shape[0] > image.shape[0]:
        sidebar = sidebar[: image.shape[0], :, :]
    elif sidebar.shape[0] < image.shape[0]:
        sidebar = np.concatenate(
            [
                np.zeros(
                    (
                        image.shape[0] - sidebar.shape[0],
                        sidebar.shape[1],
                        sidebar.shape[2],
                    ),
                    dtype=sidebar.dtype,
                ),
                sidebar,
            ],
            axis=0,
        )

    return np.concatenate([sidebar, image], axis=1)


def add_signature(
    image: np.ndarray,
    header: List[str],
    footer: List[str] = [],
    word_wrap: bool = True,
    line_width: int = 80,
) -> np.ndarray:
    if image is None or not image.shape:
        return image

    word_wrapper = lambda content: [
        line
        for line in reduce(
            lambda x, y: x + y,
            [
                (
                    [
                        " ".join(part)
                        for part in np.array_split(
                            line.split(" "),
                            int(math.ceil(len(line) / line_width)),
                        )
                    ]
                    if len(line) > 2 * line_width
                    else [line]
                )
                for line in content
            ],
            [],
        )
        if line
    ]

    if word_wrap:
        header = word_wrapper(header)
        footer = word_wrapper(footer)

    adjust_length = lambda line: (
        line if len(line) >= line_width else line + (line_width - len(line)) * " "
    )

    return np.concatenate(
        [
            render_text(
                text=adjust_length(line),
                image_width=image.shape[1],
                color_depth=image.shape[2],
            )
            for line in header
        ]
        + [image]
        + [
            render_text(
                text=adjust_length(line),
                image_width=image.shape[1],
                color_depth=image.shape[2],
            )
            for line in footer
        ],
        axis=0,
    )



import argparse
from . import *
from abcli.logger import logger


parser = argparse.ArgumentParser(NAME)
parser.add_argument(
    "task",
    type=str,
    help="status|validate",
)
args = parser.parse_args()

success = False
if args.task == "status":
    success = True
    print(str(get_status()).lower())
elif args.task == "validate":
    success = validate()
else:
    logger.error(f"-{NAME}: {args.task}: command not found.")

if not success:
    logger.error(f"-{NAME}: {args.task}: failed.")



import argparse
from . import *
from abcli.logger import logger


parser = argparse.ArgumentParser(NAME)
parser.add_argument(
    "task",
    type=str,
    default="",
    help="list_of_external",
)
parser.add_argument(
    "--delim",
    type=str,
    default=", ",
)
parser.add_argument(
    "--log",
    default=1,
    type=int,
    help="0|1",
)
parser.add_argument(
    "--repo_names",
    default=0,
    type=int,
    help="0|1",
)
args = parser.parse_args()

delim = " " if args.delim == "space" else args.delim

success = False
if args.task == "list_of_external":
    output = list_of_external(args.repo_names)
    if args.log:
        logger.info(f"{len(output):,} external plugin(s): {delim.join(output)}")
    else:
        print(delim.join(output))
    success = True
else:
    logger.error(f"-{NAME}: {args.task}: command not found.")

if not success:
    logger.error(f"-{NAME}: {args.task}: failed.")



import json
from abcli.file.classes import JsonEncoder
from abcli import string
from abcli.logger import logger, crash_report


def prepare_for_saving(filename):
    """prepare for saving filename.

    Args:
        filename (str): filename.

    Returns:
        bool: success.
    """
    from abcli.path import create as path_create
    from abcli.file.functions import path

    return path_create(path(filename))


def save(filename, data):
    """save data.

    Args:
        filename (str): filename.
        data (Any): data.

    Returns:
        bool: success.
    """
    if not prepare_for_saving(filename):
        return False

    try:
        import dill

        with open(filename, "wb") as fp:
            dill.dump(data, fp)
    except:
        crash_report(f"save({filename}) failed.")
        return False

    return True


def save_csv(
    filename,
    df,
    log=False,
):
    """save data to filename as csv.

    Args:
        filename (str): filename.
        df (pd.DataFrame): dataframe.
        log (bool, optional): log. Defaults to False.

    Returns:
        bool: success.
    """
    if not prepare_for_saving(filename):
        return False

    # https://stackoverflow.com/a/10250924/10917551
    try:
        df.to_csv(filename)
    except:
        crash_report(f"save_csv({filename}) failed.")
        return False

    if log:
        logger.info(
            "save_csv: {}X[{}] -> {}".format(
                len(df),
                ",".join(list(df.columns)),
                filename,
            )
        )

    return True


def save_fig(
    filename,
    log=False,
):
    """save plt figure to filename.

    Args:
        filename (str): filename.
        log (bool, optional): log. Defaults to False.

    Returns:
        bool: success.
    """
    from abcli.modules.host import is_jupyter

    if not prepare_for_saving(filename):
        return False

    # https://stackoverflow.com/a/10250924/10917551
    try:
        import matplotlib.pyplot as plt

        if is_jupyter():
            plt.show()
        plt.savefig(filename, bbox_inches="tight")
        plt.close()
    except:
        crash_report(f"save_fig({filename}) failed.")
        return False

    if log:
        logger.info(f"figure -> {filename}")

    return True


def save_geojson(
    filename,
    gdf,
    log=False,
):
    """save GeoDataFrame to filename as geojson.

    Args:
        filename (str): filename.
        gdf (GeoDataFrame): geo data frame.
        log (bool, optional): log. Defaults to False.

    Returns:
        bool: success.
    """
    if not prepare_for_saving(filename):
        return False

    try:
        gdf.to_file(filename, driver="GeoJSON")
    except:
        crash_report(f"save_geojson({filename}) failed.")
        return False

    if log:
        logger.info(f"save_geojson: {len(gdf)} -> {filename}")

    return True


def save_image(
    filename,
    image,
    log=False,
):
    """save image to filename.

    Args:
        filename (str): filename.
        image (np.ndarray): image.
        log (bool, optional): log. Defaults to False.

    Returns:
        bool: success.
    """
    import cv2
    import numpy as np

    if image is None:
        logger.info("save_image(None) ignored.")
        return True

    if not prepare_for_saving(filename):
        return False

    try:
        data = image.copy()

        if len(data.shape) == 3:
            data = np.flip(data, axis=2)

        cv2.imwrite(filename, data)
    except:
        crash_report(
            "save_image({},{}) failed.".format(
                string.pretty_shape_of_matrix(image),
                filename,
            )
        )
        return False

    if log:
        logger.info(
            "save_image: {} -> {}".format(
                string.pretty_shape_of_matrix(image),
                filename,
            )
        )

    return True


def save_json(
    filename,
    data,
    log=False,
):
    """save data to filename as json.

    Args:
        filename (str): filename.
        data (Any): data.
        log (bool, optional): log. Defaults to False.

    Returns:
        bool: success.
    """
    if not prepare_for_saving(filename):
        return False

    if hasattr(data, "to_json"):
        data = data.to_json()

    try:
        with open(filename, "w") as fh:
            json.dump(
                data,
                fh,
                sort_keys=True,
                cls=JsonEncoder,
                indent=4,
                ensure_ascii=False,
            )
    except:
        crash_report(f"save_json({filename}) failed.")
        return False

    if log:
        logger.info(f"save_json -> {filename}")

    return True


def save_tensor(
    filename,
    tensor,
    log=True,
):
    """save tensor to filename.

    Args:
        filename (str): filename.
        tensor (np.ndarray): tensor.
        log (bool, optional): log. Defaults to True.

    Returns:
        bool: success.
    """
    from abcli.file.functions import set_extension

    success = save(set_extension(filename, "pyndarray"), tensor)

    if success and log:
        logger.info(
            "save_tensor: {} -> {}".format(
                string.pretty_shape_of_matrix(tensor),
                filename,
            )
        )

    return success


def save_text(
    filename,
    text,
    if_different=False,
    log=False,
    remove_empty_lines=False,
):
    """save text to filename.

    Args:
        filename (str): filename.
        text (List[str]): text.
        log (bool, optional): log. Defaults to False.
        if_different (bool, optional): save if text is different
        from current filename content. Defaults to False.
        remove_empty_lines (bool, optional): remove empty lines. Defaults to False.

    Returns:
        bool: success.
    """
    from abcli.file.load import load_text

    if remove_empty_lines:
        text = [
            line
            for line, next_line in zip(text, text[1:] + ["x"])
            if line.strip() or next_line.strip()
        ]

    if if_different:
        _, content = load_text(filename, civilized=True)

        if "|".join([line for line in content if line]) == "|".join(
            [line for line in text if line]
        ):
            return True

    if not prepare_for_saving(filename):
        return False

    try:
        with open(filename, "w") as fp:
            fp.writelines([string + "\n" for string in text])
    except:
        crash_report(f"save_text({filename}) failed.")
        return False

    if log:
        logger.info(f"save_text: {len(text)} line(s) -> {filename}")
    return True


def save_yaml(
    filename,
    data,
    log=True,
):
    """save data to filename as yaml(.

    Args:
        filename (str): filename.
        data (Any): data.
        log (bool, optional): log. Defaults to True.

    Returns:
        bool: success.
    """
    if not prepare_for_saving(filename):
        return False

    try:
        import yaml

        with open(filename, "w") as f:
            yaml.dump(data, f)
    except:
        crash_report(f"save_yaml({filename}) failed.")
        return False

    if log:
        logger.info(f"save_yaml: {', '.join(data.keys())} -> {filename}.")

    return True
